/*
   File: nonterminals.c
   Maintains the administration of all lexicon/fact nonterminals with
   their formal arguments (lexicon nonterminal declarations) as
   well as all occurences of a nonterminal with actuals i.e.
   the lexicon calls occurring in a lexicon plus the fact calls.

   Also the grammar terminals are maintained here as a special kind
   of lexicon nonterminals.

   Copyright 2009-2010 Radboud University of Nijmegen

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id$"
*/

/* system includes */
#include <stdio.h>
#include <string.h>

/* libabase includes */
#include <abase_repr.h>
#include <abase_error.h>
#include <abase_memalloc.h>
#include <abase_fileutil.h>

/* local includes */
#include "options.h"
#include "dyn_array.h"
#include "affix_values.h"
#include "nonterminals.h"

/*
   We have to maintain an administration of all nonterminals with
   their arity and formal parameters in such a way, that each
   nonterminal is stored only once.

   The basic administration will therefore use hashing techniques for lookup.
   Once a nonterminal has been stored, its index in the administration
   is used as its identification.
*/
typedef struct nonterminal_rec
{ char *nont_name;
  int nont_nr;
  int fact_nr;			/* When >= 0, it is a fact */
  int_array formals;		/* The total set of formals */
  int_array criticals;		/* indicating which are critical */
} *nonterminal;

typedef struct terminal_rec
{ char *term_text;
  int marker;
} *terminal;

typedef struct call_rec
{ int nont_id;
  int_array actuals;
} *call;

static int_array all_facts;
static ptr_array all_nonterminals;
static ptr_array all_terminals;
static ptr_array all_calls;
static hash_container nonterminal_hash_container;
static hash_container call_hash_container;

#define streq(s1,s2) (strcmp((s1),(s2)) == 0)
static int hash_nonterminal (char *name, int arity, int hsize)
{ unsigned value = (unsigned) arity;
  unsigned result;
  char *ptr;

  for (ptr = name; *ptr; ptr++) value = (127 * value + ((unsigned int) (*ptr))) & 0xffffff;
  result = value % ((unsigned) hsize);
  return ((int) result);
};

int count_criticals (int_array criticals)
{ int sum = 0;
  int ix;
  for (ix = 0; ix < criticals -> size; ix++)
    if (criticals -> array[ix]) sum++;
  return (sum);
}

static int hash_call (int nont_id, int_array criticals, int *actuals, int hsize)
{ unsigned value = (unsigned) nont_id;
  unsigned result;
  int ix;

  for (ix = 0; ix < criticals -> size; ix++)
    if (!criticals -> array[ix])
      value = (997 * value + ((unsigned) actuals[ix])) & 0x1fffff;
  result = value % ((unsigned) hsize);
  return ((int) result);
};

int register_new_nonterminal (int crit, char *nont_name, int nont_nr,
			      int_array formals, int_array criticals)
{ int hindex = hash_nonterminal (nont_name, formals -> size, affix_hash_size);
  hash_list hl = nonterminal_hash_container[hindex];
  nonterminal new_nont;
  int nont_index, ix;

#ifdef DEBUG
  if (formals -> size != criticals -> size)
    abs_bug ("register_new_nonterminal", "Mismatch between formals and criticals");
#endif

  /* Locate nonterminal on hash list */
  for (ix = 0; ix < hl -> size; ix++)
    { nonterminal nont;
      int arity;
      nont_index = hl -> array[ix];
      nont = (nonterminal) (all_nonterminals -> array[nont_index]);
      arity = nont -> formals -> size;
      if (streq (nont_name, nont -> nont_name) && (formals -> size == arity))
	return (-1);	/* This should not happen */
    };

  /* Nonterminal + arity combination is new: create a unique copy */
  nont_index = all_nonterminals -> size;
  new_nont = (nonterminal) abs_malloc (sizeof (struct nonterminal_rec), "register_new_nonterminal");
  new_nont -> nont_name = abs_new_string (nont_name, "register_new_nonterminal");
  new_nont -> nont_nr = nont_nr;	/* Coder nonterminal nr */
  if (crit)
    { new_nont -> fact_nr = all_facts -> size;
      app_int_array (all_facts, nont_index);
    }
  else new_nont -> fact_nr = -1;
  new_nont -> formals = formals;
  new_nont -> criticals = criticals;
  app_ptr_array (all_nonterminals, (void *) new_nont);
  app_hash_list (hl, nont_index);
  return (nont_index);
}

int register_new_terminal (char *text, int marker)
{ terminal new_term;
  int ix;
  for (ix = 0; ix < all_terminals -> size; ix++)
    { terminal curr_term = (terminal) all_terminals -> array[ix];
      if ((strcmp (text, curr_term -> term_text) == 0) &&
	  (marker == curr_term -> marker))
	return (-1);	/* bug in the coder */
    };

  /* Terminal + marker is new: create a unique copy */
  new_term = (terminal) abs_malloc (sizeof (struct terminal_rec), "register_new_terminal");
  new_term -> term_text = abs_new_string (text, "register_new_terminal");
  new_term -> marker = marker;
  app_ptr_array (all_terminals, (void *) new_term);
  return (all_terminals -> size - 1);
}

void get_terminal_text_and_marker (int ix, char **term_text, int *marker)
{ terminal curr_term = (terminal) all_terminals -> array[ix];
  *term_text = curr_term -> term_text;
  *marker = curr_term -> marker;
}

int register_new_call (int nont_id, int_array criticals, int *actuals)
{ int hindex = hash_call (nont_id, criticals, actuals, param_hash_size);
  int nr_non_crits = criticals -> size - count_criticals (criticals);
  hash_list hl = call_hash_container[hindex];
  int call_index, ix;
  call new_call;

  /* Locate call on hash list */
  for (ix = 0; ix < hl -> size; ix++)
    { int_array call_actuals;
      int equal, iy, ca_idx;
      call this_call;

      /* Check if this call equals the one we seek */
      call_index = hl -> array[ix];
      this_call = (call) all_calls -> array[call_index];
      if (nont_id != this_call -> nont_id) continue;
      call_actuals = this_call -> actuals;

#ifdef DEBUG
      if (call_actuals -> size != nr_non_crits)
	abs_bug ("register_new_call",
		 "Arity mismatch for nont %d between registered %d and actual non critical %d",
		  nont_id, call_actuals -> size, nr_non_crits);
#endif

      /* Nonterminals id match, now check the non critical actuals */
      equal = 1;
      ca_idx = 0;
      for (iy = 0; equal && (iy < criticals -> size); iy++)
        if (!criticals -> array[iy])
	  { if (call_actuals -> array[ca_idx] != actuals[iy]) equal = 0;
	    else ca_idx++;
	  };

      /* We found a call with the same non critical actual affixes */
      if (equal) return (call_index);
    };

  /* Nonterminal id + non critical actuals combination is new: create a new call */
  call_index = all_calls -> size;
  new_call = (call) abs_malloc (sizeof (struct call_rec), "register_new_call");
  new_call -> nont_id = nont_id;
  new_call -> actuals = init_int_array (nr_non_crits);
  for (ix = 0; ix < criticals -> size; ix++)
    if (!criticals -> array[ix])
      app_int_array (new_call -> actuals, actuals[ix]);
  app_ptr_array (all_calls, (void *) new_call);
  app_hash_list (hl, call_index);
  return (call_index);
}

/*
   Lookup of nonterminals
*/
int lookup_nonterminal (char *nont_name, int arity)
{ int hindex = hash_nonterminal (nont_name, arity, affix_hash_size);
  hash_list hl = nonterminal_hash_container[hindex];
  int ix;
  for (ix = 0; ix < hl -> size; ix++)
    { int nont_index = hl -> array[ix];
      nonterminal nont = (nonterminal) (all_nonterminals -> array[nont_index]);
      if (streq (nont_name, nont -> nont_name) && (arity == nont -> formals -> size))
	return (nont_index);
    };

  /* Not found */
  return (-1);
}

int fact_nr_from_nonterminal (int nont_index)
{ nonterminal nont = (nonterminal) (all_nonterminals -> array[nont_index]);
  return (nont -> fact_nr);
}

int_array formals_from_nonterminal (int nont_index)
{ nonterminal nont = (nonterminal) (all_nonterminals -> array[nont_index]);
  return (nont -> formals);
}

int_array criticals_from_nonterminal (int nont_index)
{ nonterminal nont = (nonterminal) (all_nonterminals -> array[nont_index]);
  return (nont -> criticals);
}

int nr_of_nonterminals ()
{ return (all_nonterminals -> size);
}

int nr_of_terminals ()
{ return (all_terminals -> size);
}

int nr_of_calls ()
{ return (all_calls -> size);
}

int nr_of_facts ()
{ return (all_facts -> size);
}

int nont_id_from_fact_nr (int fact_nr)
{ return (all_facts -> array[fact_nr]);
}

/*
  Generate info file
*/
void generate_nonterminals_info (FILE *info)
{ int ix;
  fprintf (info, "Nonterminal table:\n");
  for (ix = 0; ix < all_nonterminals -> size; ix++)
    { nonterminal nont = (nonterminal) (all_nonterminals -> array[ix]);
      int arity = nont -> formals -> size;
      fprintf (info, "%4d: [%d] %s/%d", ix, nont -> nont_nr, nont -> nont_name, arity);
      if (arity)
	{ int iy;
	  fprintf (info, " (");
	  for (iy = 0; iy < arity; iy++)
	    { if (iy) fprintf (info, ", ");
	      if (nont -> criticals -> array[iy]) fprintf (info, ">");
	      fprintf (info, "%d", nont -> formals -> array[iy]);
	    };
	  fprintf (info, ")");
        };
      fprintf (info, "\n");
    };
}

void generate_terminals_info (FILE *info)
{ int ix;
  fprintf (info, "Terminal table:\n");
  for (ix = 0; ix < all_terminals -> size; ix++)
    { terminal term = (terminal) (all_terminals -> array[ix]);
      fprintf (info, "%4d: [%d], Text: '%s', Marker '%d'\n", ix,
	       ix + all_nonterminals -> size, term -> term_text, term -> marker);
    }
}

void generate_calls_info (FILE *info)
{ int ix;
  fprintf (info, "Call table:\n");
  for (ix = 0; ix < all_calls -> size; ix++)
    { call the_call = (call) (all_calls -> array[ix]);
      int nont_id = the_call -> nont_id;
      int_array actuals = the_call -> actuals;
      fprintf (info, "%4d: %d", ix, nont_id);
      if (actuals -> size)
        { int iy;
          fprintf (info, " (");
          for (iy = 0; iy < actuals -> size; iy++)
            { if (iy) fprintf (info, ", ");
              fprintf (info, "%d", actuals -> array[iy]);
            };
          fprintf (info, ")");
        };
      fprintf (info, "\n");
    };
}

/*
   Dumping of information
*/
void dump_nonterminal (FILE *dump, int nont_id)
{ nonterminal nont = (nonterminal) (all_nonterminals -> array[nont_id]);
  int_array formals = nont -> formals;
  int_array criticals = nont -> criticals;
  int ix;
  fprintf (dump, "%s", nont -> nont_name);
  if (!formals -> size) return;
  fprintf (dump, " (");
  for (ix = 0; ix < formals -> size; ix++)
    { if (ix) fprintf (dump, ", ");
      if (criticals -> array[ix]) fprintf (dump, ">");
      fprintf (dump, "%s", affix_name_from_index (formals -> array[ix]));
    };
  fprintf (dump, ")");
}

void dump_fact (FILE *dump, int fact_nr)
{ if (fact_nr == nr_of_facts ())
    fprintf (dump, "-- Triple table --");
  else dump_nonterminal (dump, all_facts -> array[fact_nr]);
}

void dump_call (FILE *dump, int call_id)
{ call the_call = (call) (all_calls -> array[call_id]);
  int_array fpars, criticals, actuals;
  int nont_id = the_call -> nont_id;
  nonterminal nont;
  int cidx, ix;
  if (nont_id >= all_nonterminals -> size)
    { terminal term = (terminal) (all_terminals -> array[nont_id - all_nonterminals -> size]);
      fprintf (dump, "\"%s\" [%d]", term -> term_text, term -> marker);
      return;
    };
  nont = (nonterminal) (all_nonterminals -> array[nont_id]);
  fpars = nont -> formals;
  criticals = nont -> criticals;
  actuals = the_call -> actuals;
  fprintf (dump, "%s", nont -> nont_name);
  if (!fpars -> size) return;
  fprintf (dump, " (");
  for (ix = 0, cidx = 0; ix < fpars -> size; ix++)
    { if (ix) fprintf (dump, ", ");
      if (criticals -> array[ix]) fprintf (dump, ">_");
      else
	{ dump_affix_value (dump, fpars -> array[ix], actuals -> array[cidx]);
	  cidx++;
        };
    };
  fprintf (dump, ")");
}

/*
   Binary save the nonterminals
   Binary verify the lexicon header against the lif
*/
void bin_save_nonterminals (BinFile bf)
{ int ix;
  abs_bin_save_int (bf, all_nonterminals -> size);
  for (ix = 0; ix < all_nonterminals -> size; ix++)
    { nonterminal nont = (nonterminal) (all_nonterminals -> array[ix]);
      abs_bin_save_string (bf, nont -> nont_name);
      abs_bin_save_int (bf, nont -> nont_nr);
      bin_save_int_array (bf, nont -> formals);
      bin_save_int_array (bf, nont -> criticals);
    };
}

void bin_save_terminals (BinFile bf)
{ int ix;
  abs_bin_save_int (bf, all_terminals -> size);
  for (ix = 0; ix < all_terminals -> size; ix++)
    { terminal term = (terminal) (all_terminals -> array[ix]);
      abs_bin_save_string (bf, term -> term_text);
      abs_bin_save_int (bf, term -> marker);
    }
}

void bin_save_calls (BinFile bf)
{ int ix, iy;
  abs_bin_save_int (bf, all_calls -> size);
  for (ix = 0; ix < all_calls -> size; ix++)
    { call the_call = (call) (all_calls -> array[ix]);
      int_array actuals = the_call -> actuals;
      abs_bin_save_int (bf, the_call -> nont_id);
      for (iy = 0; iy < actuals -> size; iy++)
	abs_bin_save_int (bf, actuals -> array[iy]);
    };
}

static int is_new_nonterminal (BinFile bf, nonterminal nont)
{ int ival, size, nont_nr, ix;
  char *sval;
  abs_bin_load_string (bf, &sval);
  if (!streq (sval, nont -> nont_name)) return (1);
  abs_free (sval, "is_new_nonterminal");
  abs_bin_load_int (bf, &nont_nr);
  if (nont_nr != nont -> nont_nr) return (1);
  abs_bin_load_int (bf, &size);
  if (size != nont -> formals -> size) return (1);
  for (ix = 0; ix < nont -> formals -> size; ix++)
    { abs_bin_load_int (bf, &ival);   
      if (ival != nont -> formals -> array[ix])
	return (1);
    };
  abs_bin_load_int (bf, &size);
  if (size != nont -> criticals -> size) return (1);
  for (ix = 0; ix < nont -> criticals -> size; ix++)
    { abs_bin_load_int (bf, &ival);
      if (ival != nont -> criticals -> array[ix])
	return (1);
    };
  return (0);
}

int has_new_lif_nonterminals (BinFile bf)
{ int ix, value;
  abs_bin_load_int (bf, &value);
  if (value != all_nonterminals -> size) return (1);
  for (ix = 0; ix < all_nonterminals -> size; ix++)
    if (is_new_nonterminal (bf, (nonterminal) (all_nonterminals -> array[ix])))
      return (1);
  return (0);
}

static int is_new_terminal (BinFile bf, terminal term)
{ char *sval;
  int ival;
  abs_bin_load_string (bf, &sval);
  if (!streq (sval, term -> term_text)) return (1);
  abs_free (sval, "is_new_terminal");
  abs_bin_load_int (bf, &ival);
  return (ival != term -> marker);
}

int has_new_lif_terminals (BinFile bf)
{ int ix, value;
  abs_bin_load_int (bf, &value);
  if (value != all_terminals -> size) return (1);
  for (ix = 0; ix < all_terminals -> size; ix++)
    if (is_new_terminal (bf, (terminal) (all_terminals -> array[ix])))
      return (1);
  return (0);
}

/*
   Initialization stuff
*/
#define MAX_NR_TERMINALS 2048
void init_nonterminals ()
{ int ix;
  all_facts = init_int_array (affix_hash_size);
  all_nonterminals = init_ptr_array (affix_hash_size);
  all_terminals = init_ptr_array (MAX_NR_TERMINALS);
  nonterminal_hash_container =
	(hash_container) abs_calloc (affix_hash_size, sizeof (hash_list), "init_nonterminals");
  for (ix = 0; ix < affix_hash_size; ix++)
    nonterminal_hash_container[ix] = new_hash_list();
  all_calls = init_ptr_array (param_hash_size);
  call_hash_container =
        (hash_container) abs_calloc (param_hash_size, sizeof (hash_list), "init_calls");
  for (ix = 0; ix < param_hash_size; ix++)
    call_hash_container[ix] = new_hash_list();
}
