/*
   File: lxcn_lexicon.c
   Maintains a lexicon and reads a lexicon as generated by agfl-lexgen

   Copyright 2009-2010 Radboud University of Nijmegen
 
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id$"
*/

/* standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* libabase includes */
#include <abase_porting.h>
#include <abase_repr.h>
#include <abase_error.h>
#include <abase_fileutil.h>
#include <abase_memalloc.h>

/* local includes */
#include "lxcn_vocabulary.h"
#include "lxcn_lexicon.h"
#include "lxcn_lexicon_impl.h"
#include "lxcn_search.h"

/*------------------------------------------------------------------------------
// Format of lexicon file (as delivered by agfl-lexgen)
//
// 0a) lexicon header
// 0b) int		nr_lexicons
// 0c) char *		lexicon_names[]
// 0d) int		nr_fact_tables
// 0e) char *		fact_table_names[]
// 0f) int		nr_triples_databases
// 0g) char *		triples_database_names[]
// 0h) int		hybrid_parsing_flag
//
// The affix table: the nr of lif affixes is saved to check for lexicon recompilation.
//
// 1a) int		nr_lif_affixes
// 1b) int      	nr_set_affixes
// 1c) setaffix		set_affixes[]
// 2a) int		nr_int_affixes
// 2b) int		int_affixes[]
// 3a) int		nr_text_affixes
// 3b) char *		text_affixes[]
//
// The nonterminal and call table:
//
// 4a) int		nr_nonterminals
// 4b) nonterminal	nonterminals[]
// 4c) int		nr_terminals
// 4d) terminal		terminals[]
// 5a) int		nr_calls
// 5b) call		calls
//
// The mapping from list nr to calls
// 6a) int		nr_entries
// 6b) entry		all_entries (saved as int *)
//
// The vocabularies
// 7a) int		nr_vocs
// 7b) int		markers
// 7c) Vocabulary	vocabularies
//
// The fact table administration
// 8a) Vocabulary       critical_texts
// 8b) int              nr_fact_tables
// 8c) FactTable        all_fact_tables[]
//----------------------------------------------------------------------------*/

/* Skip row of texts containing lexicon names */
static void bin_skip_text_array (BinFile bf)
{ int size, ix;
  char *sval;
  abs_bin_load_int (bf, &size);
  for (ix = 0; ix < size; ix++)
    { abs_bin_load_string (bf, &sval);
      abs_free (sval, "bin_skip_text_array");
    };
}

static void bin_load_set_affix (Pool p, BinFile bf, SetAffix *affix)
{ abs_bin_load_pool_string (p, bf, &affix -> name);
  abs_bin_load_int64  (bf, &affix -> bitset);
}

static void bin_load_affix_values (BinFile bf, Lexicon lex)
{ int nr_lif_affixes, ix;

  /* Load set affixes */
  abs_bin_load_int (bf, &nr_lif_affixes);	/* Ignore value */
  abs_bin_load_int (bf, &lex -> nr_set_affixes);
  lex -> set_affixes = (SetAffix *) abs_pool_malloc (lex -> pool,
			lex -> nr_set_affixes * sizeof (SetAffix), "bin_load_affix_values");
  for (ix = 0; ix < lex -> nr_set_affixes; ix++)
    bin_load_set_affix (lex -> pool, bf, &lex -> set_affixes[ix]);

  /* load int affixes */
  abs_bin_load_int (bf, &lex -> nr_int_affixes);
  lex -> int_affixes = (int *) abs_pool_malloc (lex -> pool, lex -> nr_int_affixes * sizeof (int),
					        "bin_load_affix_values");
  for (ix = 0; ix < lex -> nr_int_affixes; ix++)
    abs_bin_load_int (bf, &lex -> int_affixes[ix]);

  /* load text affixes */
  abs_bin_load_int (bf, &lex -> nr_text_affixes);
  lex -> text_affixes = (char **) abs_pool_malloc (lex -> pool,
			lex -> nr_text_affixes * sizeof (char *), "bin_load_affix_values");
  for (ix = 0; ix < lex -> nr_text_affixes; ix++)
    abs_bin_load_pool_string (lex -> pool, bf, &lex -> text_affixes[ix]);
}

static void bin_load_nonterminal (BinFile bf, Nonterminal *nont, Lexicon lex)
{ int ix, npars;
  abs_bin_load_pool_string (lex->pool, bf, &nont -> name);
  abs_bin_load_int (bf, &nont -> nont_nr);
  abs_bin_load_int (bf, &nont -> arity);
  nont -> ncrits = 0;
  nont -> fpars = (int *) abs_pool_malloc (lex -> pool,
					   nont -> arity * sizeof (int), "bin_load_nonterminal");
  nont -> crits = (int *) abs_pool_malloc (lex -> pool,
					   nont -> arity * sizeof (int), "bin_load_nonterminal");
  for (ix = 0; ix < nont -> arity; ix++)
    abs_bin_load_int (bf, &nont -> fpars[ix]);
  abs_bin_load_int (bf, &npars);
  if (npars != nont -> arity)
    abs_bug ("bin_load_nonterminal", "Mismatch between formals and criticals");
  for (ix = 0; ix < nont -> arity; ix++)
    { abs_bin_load_int (bf, &nont -> crits[ix]);
      if (nont -> crits[ix]) nont -> ncrits++;
    };
}

static void bin_load_nonterminals (BinFile bf, Lexicon lex)
{ int ix;
  abs_bin_load_int (bf, &lex -> nr_nonterminals);
  lex -> nonterminals = (Nonterminal *) abs_pool_malloc (lex -> pool,
			lex -> nr_nonterminals * sizeof (Nonterminal), "bin_load_nonterminals");
  for (ix = 0; ix < lex -> nr_nonterminals; ix++)
    bin_load_nonterminal (bf, &lex -> nonterminals[ix], lex);
}

static void bin_skip_terminal (BinFile bf)
{ char *sval;
  int marker;
  abs_bin_load_string (bf, &sval);
  abs_free (sval, "bin_skip_terminal");
  abs_bin_load_int (bf, &marker);
}

static void bin_load_terminals (BinFile bf, Lexicon lex)
{ int ix;
  abs_bin_load_int (bf, &lex -> nr_terminals);
  for (ix = 0; ix < lex -> nr_terminals; ix++)
    bin_skip_terminal (bf);
}

char *lxcn_get_nonterminal_name (Lexicon lex, int nont_nr)
{ Nonterminal *nont = &lex -> nonterminals[nont_nr];
  return (nont -> name);
}

static void bin_load_call (BinFile bf, Call *call, Lexicon lex)
{ int ix, nr_fpars;
  Nonterminal *nont;
  abs_bin_load_int (bf, &call -> nont_id);
  if (call -> nont_id >= lex -> nr_nonterminals)
    { int term_id = call -> nont_id - lex -> nr_nonterminals;
      if (term_id >= lex -> nr_terminals)
 	abs_bug ("bin_load_call", "Loading illegal terminal");
      call -> actuals = NULL;
      return;
    };
  nont = &lex -> nonterminals[call -> nont_id];
  nr_fpars = nont -> arity - nont -> ncrits;
  call -> actuals = (int *) abs_pool_malloc (lex -> pool, nr_fpars * sizeof (int), "bin_load_call");
  for (ix = 0; ix < nr_fpars; ix++)
    abs_bin_load_int (bf, &(call -> actuals[ix]));
}

static void bin_load_calls (BinFile bf, Lexicon lex)
{ int ix;
  abs_bin_load_int (bf, &lex -> nr_calls);
  lex -> calls = (Call *) abs_pool_malloc (lex -> pool,
					   lex -> nr_calls * sizeof (Call), "bin_load_calls");
  for (ix = 0; ix < lex -> nr_calls; ix++)
    bin_load_call (bf, &lex -> calls[ix], lex);
}

static void bin_load_entries (BinFile bf, Lexicon lex)
{ int ix;
  abs_bin_load_int (bf, &lex -> nr_lex_entries);
  lex -> all_entries = (int **) abs_pool_malloc (lex -> pool,
		        	lex -> nr_lex_entries * sizeof (int *), "bin_load_entries");
  lex -> all_entries[0] = NULL;
  for (ix = 1; ix < lex -> nr_lex_entries; ix++)
    { int nr_in_list, iy;
      int *list;
      abs_bin_load_int (bf, &nr_in_list);
      list = (int *) abs_pool_malloc (lex -> pool,
				      (2 * nr_in_list + 1) * sizeof (int), "bin_load_entries");
      lex -> all_entries[ix] = list;
      list[0] = nr_in_list;
      for (iy = 0; iy < nr_in_list; iy++)
	{ abs_bin_load_int (bf, &list [2 * iy + 1]);	/* call_id */
	  abs_bin_load_int (bf, &list [2 * iy + 2]);	/* frequency */
	};
    };
}

static void bin_load_vocabularies (BinFile bf, Lexicon lex)
{ int ix;
  abs_bin_load_int (bf, &lex -> nr_vocabularies);
  lex -> all_lexeme_markers = (int *) abs_pool_calloc (lex -> pool,
			      lex -> nr_vocabularies, sizeof (int), "bin_load_vocabularies");
  lex -> all_vocabularies = (Vocabulary *) abs_pool_calloc (lex -> pool,
			    lex -> nr_vocabularies, sizeof (Vocabulary), "bin_load_vocabularies");
  for (ix = 0; ix < lex -> nr_vocabularies; ix++)
    { abs_bin_load_int (bf, &lex -> all_lexeme_markers[ix]);
      lxcn_bin_pool_load_trie (lex->pool, bf, &lex -> all_vocabularies[ix]);
    };
}

static void bin_load_fact_tables (BinFile bf, Lexicon lex)
{ int act_nr_facts;
  int ix;
  lxcn_bin_pool_load_trie (lex -> pool, bf, &lex -> all_critical_texts);
  abs_bin_load_int (bf, &lex -> nr_facts);
  act_nr_facts = lex -> nr_facts + lex -> hybrid_parsing;
  lex -> all_fact_tables = (FactTable *) abs_pool_calloc (lex -> pool,
			   act_nr_facts, sizeof (FactTable), "bin_load_fact_tables");
  for (ix = 0; ix < act_nr_facts; ix++)
    lxcn_bin_load_fact_table (lex -> pool, bf, &lex -> all_fact_tables[ix]);
}

Lexicon lxcn_read_lexicon (char *path)
{ Pool pool = abs_pool_init (1 * 1024 * 1024L, 8 * 1024 * 1024L);
  Lexicon lex = (Lexicon) abs_pool_malloc (pool, sizeof (struct lexicon_rec), "lxcn_read_lexicon");
  BinFile bf = abs_bin_fopen (path, "r");
  int err = 0;

  lex -> pool = pool;
  lex -> all_critical_texts = NULL;
  if (bf == NULL) 
    err = -1;
  else
    { /*lex -> stringpool = abs_pool_init (1 * 1024 * 1024L, 8 * 1024 * 1024L); */
      err = abs_bin_verify_version (bf, "binary lexicon", 0);
      if (err == 0)
	{ bin_skip_text_array (bf);			 /* Lexicon names */
	  bin_skip_text_array (bf);			 /* Fact table names */
	  bin_skip_text_array (bf);			 /* Triple database names */
	  abs_bin_load_int (bf, &lex -> hybrid_parsing); /* Hybrid parsing flag */
	  bin_load_affix_values (bf, lex);
	  bin_load_nonterminals (bf, lex);
	  bin_load_terminals (bf, lex);
	  bin_load_calls (bf, lex);
	  bin_load_entries (bf, lex);
	  bin_load_vocabularies (bf, lex);
	  bin_load_fact_tables (bf, lex);
	  abs_bin_verify_eof (bf);
	};
      abs_bin_fclose (bf);
    };

  /* abs_pool_stats(pool, "lxcn_read_lexicon"); */
  if (err != 0)
    { lxcn_free_lexicon(lex);
      lex = NULL;
    };
  return (lex);
}

void lxcn_free_lexicon (Lexicon lex)
{ /* abs_pool_free (lex->stringpool, "lxcn_free_lexicon string"); */
  abs_pool_free (lex -> pool, "lxcn_free_lexicon block");
  lxcn_dispose_garbage ();
}

/*
   Access and other methods on the Lexicon structure
*/
int lxcn_get_entries_from_nr (Lexicon lex, int entry_nr, int **entries)
{ int *list = lex -> all_entries[entry_nr];
  *entries = list;
  return (list[0]);
}

void lxcn_get_entry_params (Lexicon lex, int *entries, int idx, int *kind, int *ret_nr,
			    int *ret_arity, int *ret_freq, Value **ret_params)
{ int call_id, nont_id, nont_nr, arity, ix;
  int *fpars, *actuals;
  Nonterminal nont;
  Value *params;
  Call call;
  int cidx;

  /* Check for a valid index */
  if ((idx < 0) || (idx >= *entries))
    abs_bug ("lxcn_get_entry_params", "Illegal entry index %d", idx);

  /* Pick up the call and frequency */
  call_id = entries[2 * idx + 1];
  *ret_freq = entries[2 * idx + 2];
  call = lex -> calls[call_id]; 

  /* Pick up (non)terminal nr */
  nont_id = call.nont_id;
  if (nont_id >= lex -> nr_nonterminals)
    { /* entry is a grammar terminal */
      *ret_nr = nont_id - lex -> nr_nonterminals;
      *ret_arity = 0;
      *ret_params = NULL;
      *kind = 1;
      return;
    };

  /* Pick up arity and formal args */
  actuals = call.actuals;
  nont = lex -> nonterminals[nont_id];
  nont_nr = nont.nont_nr;
  arity = nont.arity - nont.ncrits;
  fpars = nont.fpars;

  /* Note that we do not use specifically tagged values */
  params = (Value *) abs_calloc (arity, sizeof (Value), "lxcn_get_entry_params");
  
  /* Fill in actual args, depending on formal argument type */
  cidx = 0;
  for (ix = 0; ix < nont.arity; ix++)
    if (!nont.crits[ix])
      { switch (fpars[ix])
          { case FormalINT:  params[cidx].int_par  = lex -> int_affixes[actuals[cidx]]; break;
	    case FormalTEXT: params[cidx].text_par = lex -> text_affixes[actuals[cidx]]; break;
	    default: /* Set affix */
	      params[cidx].set_par = lex -> set_affixes[actuals[cidx]].bitset;
          };
	cidx++;
      };

  /* Return the santenkraam */
  *ret_nr = nont_nr;
  *ret_arity = arity;
  *ret_params = params;
  *kind = 0;
}

void lxcn_print_entry (Lexicon lex, int *entries, int idx)
{ int call_id, freq, nont_id, arity, ix;
  int *fpars, *actuals;
  Nonterminal nont;
  Call call;

  /* Check for a valid index */
  if ((idx < 0) || (idx >= *entries))
    abs_bug ("lxcn_get_entry_params", "Illegal entry index %d", idx);

  /* Pick up the call and frequency */
  call_id = entries[2 * idx + 1];
  freq = entries[2 * idx + 2];
  call = lex -> calls[call_id]; 

  /* Pick up nonterminal nr */
  nont_id = call.nont_id;
  actuals = call.actuals;
  if (nont_id >= lex -> nr_nonterminals)
    { abs_printf ("Syntax terminal %d\n", nont_id - lex -> nr_nonterminals);
      return;
    };

  /* nonterminal name, arity and formal args */
  nont = lex -> nonterminals[nont_id];
  abs_printf ("%s", nont.name);
  arity = nont.arity - nont.ncrits;
  fpars = nont.fpars;

  /* Print the santenkraam */
  if (arity)
    { int cidx = 0;
      abs_printf (" (");
      for (ix = 0; ix < nont.arity; ix++)
        if (!nont.crits[ix])
          { if (cidx) abs_printf (", ");
	    switch (fpars[ix])
	      { case FormalINT:  abs_printf ("%d", lex -> int_affixes[actuals[cidx]]); break;
	        case FormalTEXT: abs_printf ("\"%s\"", lex -> text_affixes[actuals[cidx]]); break;
	        default: /* Set affix */
	          abs_printf ("%s", lex -> set_affixes[actuals[cidx]].name);
	      };
	  };
      abs_printf (")");
    };
  abs_printf ("\n");
}
